package ch.zhaw.ads;

import java.util.*;

public class TreeTraversal<T extends Comparable<T>> implements Traversal<T> {
    @SuppressWarnings("unchecked")
    private final TreeNode<T> root;

    public TreeTraversal(TreeNode<T> root) {
        this.root = root;
    }


    @Override
    public void interval(T min, T max, Visitor<T> vis) {
        Queue<TreeNode<T>> queue =  new LinkedList<TreeNode<T>>();
        queue.add(root);
        while(!queue.isEmpty()){
            TreeNode<T> node = queue.poll();
            if(node.getValue().compareTo(min) >= 0 && node.getValue().compareTo(max) <= 0)
            vis.visit(node.getValue());
            if(node.left != null && (node.right == null || node.right.getValue().compareTo(min) >= 0)){
                queue.add(node.left);
            }
            if(node.right != null && (node.left == null || node.left.getValue().compareTo(max) <= 0)){
                queue.add(node.right);
            }

        }
    }

    public void inorder(Visitor<T> vis) {
        inorderRec(vis, root);
    }

    private void inorderRec(Visitor<T> vis, TreeNode<T> root){
        if(root.left != null){
            inorderRec(vis, root.left);
        }
        vis.visit(root.getValue());
        if(root.right != null){
            inorderRec(vis, root.right);
        }
    }

    public void preorder(Visitor<T> vis) {
        Stack<TreeNode<T>> stack = new Stack<TreeNode<T>>();
        TreeNode<T> node = root;
        stack.push(node);
        while(!stack.isEmpty()) {
            node = stack.pop();
            vis.visit(node.getValue());
            if(node.right != null){
                stack.push(node.right);
            }
            if(node.left != null) {
                stack.push(node.left);
            }
        }
    }

    public void postorder(Visitor<T> vis) {
        Stack<TreeNode<T>> stack = new Stack<TreeNode<T>>();
        TreeNode<T> root = this.root;
        do {
            while(root != null){
                if(root.right != null)  {
                    stack.push(root.right);
                }
                stack.push(root);
                root = root.left;
            }
            root = stack.pop();
            if(root.right != null && !stack.isEmpty() && root.right == stack.peek()){
                stack.pop();
                stack.push(root);
                root = root.right;
            } else {
                vis.visit(root.getValue());
                root = null;
            }
        } while(!stack.isEmpty());
    }

    public void levelorder(Visitor<T> vis){
        Queue<TreeNode<T>> queue =  new LinkedList<TreeNode<T>>();
        queue.add(root);
        while(!queue.isEmpty()){
            TreeNode<T> node = queue.poll();
            vis.visit(node.getValue());
            if(node.left != null){
                queue.add(node.left);
            }
            if(node.right != null){
                queue.add(node.right);
            }

        }
    }

}
